
-------------------------------------------------------------------------
-- Util :: ファイル全体を文字列として読み込み
-------------------------------------------------------------------------

function readFile(filename)
  local fh   = io.open(filename,"rb")
  local data = fh:read("*a")
  fh:close()
  return data
end

-------------------------------------------------------------------------
-- Load :: luacの出力をロード
--           - リトルエンディアン
--           - sizeof(int)==sizeof(size_t)==sizeof(Instruction)==4
--           - sizeof(double)==8
--           - lua_Number == double
--         を仮定しています
--
-- 関数オブジェクト(Proto)の表現はこんな感じ
--   {
--     numup  = upvalueの数
--     numpr  = 引数の数
--     varflg = 可変長引数フラグ
--     maxstk = 最大スタック使用量
--     code   = 命令(32bit整数)の配列
--     kv     = 定数値(number,string,boolean,nil)の配列
--     kf     = 内部関数(proto)の配列
--   }
-- number,string,boolean,nil は Lua の値をそのまま用います
--
-- 使用例:
--   ld = loader(readFile("luac.out"))
--   main = ld:func()  -- トップレベルスクリプトの関数オブジェクト
-------------------------------------------------------------------------

function loader(d)
  -- ヘッダチェック。ほぼスキップ
  if d:sub(1,4) ~= "\027Lua" then
    error("Not a valid Lua chunk file!")
  end
  d = d:sub(13)

  return {
    d     = d,
    bytes =
      function(self, n)
        local dd = self.d
        self.d = self.d:sub(n+1)
        return dd:byte(1,n)
      end,
    -----------------------------
    byte = -- 1バイト整数
      function(self)
        return self:bytes(1)
      end,
    bool = -- 論理値
      function(self)
        return self:byte(1) ~= 0
      end,
    int  = -- 4バイト整数（リトルエンディアン）
      function(self)
        local x,y,z,w = self:bytes(4)
        return ((w*256+z)*256+y)*256+x
      end,
    string = -- 文字列 = 長さ(size_t) ; char * 長さ ; \0
      function(self)
        local n = self:int()
        if n == 0 then
          return ""
        else
          local dd = self.d
          self.d = self.d:sub(n+1)
          return dd:sub(1,n-1)
        end
      end,
    number = -- 数値 = 倍精度(1+11+52)浮動小数点数
      function(self)
        local l,h,s = self:int(), self:int(), -1
        if h < 2^31 then s = 1 end
        local e = 1 + math.floor(h/2^20)%2^11 - 2^10
        if e == -1023 then
          return 0
        else
          return s*(2^e + math.ldexp(h%2^20,e-20) + math.ldexp(l,e-52))
        end
      end,
    code = -- コード = 4バイト整数の列
      function(self)
        local n = self:int()
        local code = {}
        for i=1,n do
          code[i] = self:int()
        end
        return code
      end,
    func = -- 関数
      function(self)
        local obj = {}
        self:string() -- 定義されたソースファイル名
        self:int()    -- 開始行
        self:int()    -- 終了行
        obj.numup = self:byte()
        obj.numpr = self:byte()
        obj.varflg= self:byte()
        obj.maxstk= self:byte()
        obj.code  = self:code()
        obj.kv    = self:constsV()
        obj.kf    = self:constsF()
        self:dbginfo()
        return obj
      end,
    constsV = -- 定数テーブル
      function(self)
        local n  = self:int()
        local kv = {}
        for i=1, n do
          local t = self:byte()
          if t == 0 then -- TNIL
            kv[i] = nil
          elseif t == 1 then -- TBOOLEAN
            kv[i] = self:bool()
          elseif t == 3 then -- TNUMBER
            kv[i] = self:number()
          elseif t == 4 then -- TSTRING
            kv[i] = self:string()
          else
            error("Bad Tag")
          end
        end
        return kv
      end,
    constsF = -- 内部関数テーブル
      function(self)
        local n  = self:int()
        local kf = {}
        for i=1, n do
          kf[i] = self:func()
        end
        return kf
      end,
    dbginfo = -- デバッグ情報。とりあえずスキップ
      function(self)
        local n = self:int()
        for i=1, n do
          self:int()
        end
        n = self:int()
        for i=1, n do
          self:string()
          self:int()
          self:int()
        end
        n = self:int()
        for i=1, n do
          self:string()
        end
      end,
  }
end

-------------------------------------------------------------------------
-- Util :: (n>>lsb) の下位 numbits ビット
-------------------------------------------------------------------------

function bits(n, lsb, numbits)
  return math.floor(n / 2^lsb) % 2^numbits
end

-------------------------------------------------------------------------
-- VM :: protoオブジェクトを渡すとclosureオブジェクトに変える
--  upvalue は { base = stk/self, idx = インデックス, [1]=close済み値 }
--  という表現
-------------------------------------------------------------------------

function toclosure(proto)
  -- とりあえず複製
  local cl = {}
  for k,v in pairs(proto) do
    cl[k] = v
  end

  -- Upvalueフィールド追加
  cl.upval = {}
  for i=1,cl.numup do
    cl.upval[i] = nil
  end
  return cl
end

-------------------------------------------------------------------------
-- Lib :: ライブラリ関数パッケージを返す
-------------------------------------------------------------------------

isfn = {}
function library()
  return {
     print  = print,
     string = string,
     io     = io,
     math   = math,
     os     = os,
     table  = table,
     -- debug
     -- package
     -- coroutine
     tonumber = tonumber,
     rawget   = rawget,
     rawset   = rawset,
     rawequal = rawequal,
     arg      = {[-1]=arg[-1].." "..arg[0], [0]=arg[1], unpack(arg,2)},
     pairs    = pairs,
     ipairs   = ipairs,
     error    = error,
     unpack   = unpack,
     type     = function(v)
                  local s = type(v)
                  return (s=="table" and v.isfn==isfn) and "function" or s
                end,
  }
end

-------------------------------------------------------------------------
-- VM
--   protoオブジェクトを渡すとそれをメインスクリプトとして実行
-------------------------------------------------------------------------

function vmLoop(proto)
  -- スタックとその基本操作
  local stk  = {}
  local base = 1
  local top  = base + proto.maxstk
  local retpos = 0
  local function getreg(n)   return stk[base+n] end
  local function setreg(n,v) stk[base+n] = v    end
  local function getregs(i,n) return unpack(stk, base+i, base+i+n-1) end
  local function setregs(i,n,...)
    local a = {...}
    local m = n
    if m > #a then m = #a end
    for j=1,m   do stk[base+i+j-1] = a[j] end
    for j=m+1,n do stk[base+i+j-1] = nil  end
  end

  -- 同じ変数に対しては同じUpvalueを返すための記憶テーブル
  local uvcache = {}

  function closeupvals(from) -- upvalueをスタックから逃がす処理
    from = from or 0
    for k,u in pairs(uvcache) do
      if from<=k and not rawequal(u.base, u) then
        u[1]   = u.base[u.idx]
        u.base = u
        u.idx  = 1
        uvcache[k] = nil --逃がし終わったらテーブルから消す
      end
    end
  end

  -- 呼び出し履歴スタック
  local cistk = {}
  local citop = 1
  local function cipush(...) cistk[citop]={...} citop=citop+1 end
  local function cipop()     citop=citop-1 return unpack(cistk[citop]) end

  -- グローバルテーブル
  local global_table = library()
  global_table._G = global_table

  -- メインの実行ループ
  local pc = 1
  while true do
    local inst = proto.code[pc]
    pc = pc+1

    local OP  = bits(inst, 0,6)
    local A   = bits(inst, 6,8)
    local C   = bits(inst,14,9)
    local B   = bits(inst,23,9)
    local Bx  = B*2^9 + C
    local sBx = Bx - 131071
    local RKB = getreg(B)   if B>=2^8 then RKB = proto.kv[B-2^8+1] end
    local RKC = getreg(C)   if C>=2^8 then RKC = proto.kv[C-2^8+1] end

    if     OP ==  0 then -- MOVE  : R[A] = R[B]
       setreg(A, getreg(B))
    elseif OP ==  1 then -- LOADK : R[A] = ConstValue[Bx]
       setreg(A, proto.kv[Bx+1])
    elseif OP ==  2 then -- LOADBOOL : R[A] = (bool)B ; if C pc++
       setreg(A, B~=0)
       if C~=0 then pc = pc+1 end
    elseif OP ==  3 then -- LOADNIL : R[A..B] = nil
       for i=A,B do
         setreg(i, nil)
       end
    elseif OP ==  4 then -- GETUPVAL : R[A] = upval[B]
       local u = proto.upval[B+1]
       setreg(A, u.base[u.idx])
    elseif OP ==  5 then -- GETGLOBAL : R[A] = _G[ConstValue[Bx]]
       setreg(A, rawget(global_table, proto.kv[Bx+1]))
    elseif OP ==  6 then -- GETTABLE : R[A] = RB[RKC]
       -- ToDo: Metatable...
       local rb = getreg(B)
       setreg(A, rb[RKC])
    elseif OP ==  7 then -- SETGLOBAL : _G[ConstValue[Bx]] = R[A]
       rawset(global_table, proto.kv[Bx+1], getreg(A))
    elseif OP ==  8 then -- SETUPVAL : upval[B] = R[A]
       local u = proto.upval[B+1]
       u.base[u.idx] = getreg(A)
    elseif OP ==  9 then -- SETTABLE : RA[RKB] = RKC
       local ra = getreg(A)
       ra[RKB] = RKC
    elseif OP == 10 then -- NEWTABLE : R[A] = {}  -- B,C はハッシュの予定サイズ
       setreg(A, {})
    elseif OP == 11 then -- SELF : R[A]=RB[RKC] ; R[A+1]=RB
       local rb = getreg(B)
       setreg(A,   rb[RKC])
       setreg(A+1, rb)
    elseif OP == 12 then -- ADD : R[A] = RK[B] + RK[C]
       setreg(A, RKB + RKC)
    elseif OP == 13 then -- SUB : R[A] = RK[B] - RK[C]
       setreg(A, RKB - RKC)
    elseif OP == 14 then -- MUL : R[A] = RK[B] * RK[C]
       setreg(A, RKB * RKC)
    elseif OP == 15 then -- DIV : R[A] = RK[B] / RK[C]
       setreg(A, RKB / RKC)
    elseif OP == 16 then -- MOD : R[A] = RK[B] % RK[C]
       setreg(A, RKB % RKC)
    elseif OP == 17 then -- POW : R[A] = RK[B] ^ RK[C]
       setreg(A, RKB ^ RKC)
    elseif OP == 18 then -- UNM : R[A] = - R[B] // RKBじゃないのは定数なら畳み込むから？でも手抜き
       setreg(A, - RKB)
    elseif OP == 19 then -- NOT : R[A] = not R[B] // RKBじゃないのは定数なら畳み込むから？でも手抜き
       setreg(A, not RKB)
    elseif OP == 20 then -- LEN : R[A] = # R[B] // RKBじゃないのは定数なら畳み込むから？でも手抜き
       setreg(A, # RKB)
    elseif OP == 21 then -- CONCAT : R[A] = R[B] .. R[B+1] .. … .. RK[C]
       local s = getreg(B)
       for i=B+1,C do s = s .. getreg(i) end
       setreg(A, s)
    elseif OP == 22 then -- JMP : pc += sBx
       pc = pc + sBx
    elseif OP == 23 then -- EQ : 次のjmp命令を (RKB==RKC)==A ? 実行 : スキップ
       if (RKB==RKC) ~= (A>0) then pc=pc+1 end
    elseif OP == 24 then -- LT : 次のjmp命令を (RKB< RKC)==A ? 実行 : スキップ
       if (RKB< RKC) ~= (A>0) then pc=pc+1 end
    elseif OP == 25 then -- LE : 次のjmp命令を (RKB<=RKC)==A ? 実行 : スキップ
       if (RKB<=RKC) ~= (A>0) then pc=pc+1 end
    elseif OP == 26 then -- TEST : 次のjmp命令を (RA)==C ? 実行 : スキップ
       if (not getreg(A)) == (C>0) then pc=pc+1 end
    elseif OP == 27 then -- TESTSET : 次のjmp命令を (RA)==C ? (RA=RB;実行) : スキップ
       if (not getreg(B)) == (C>0) then pc=pc+1 else setreg(A,getreg(B)) end
    elseif OP == 28 then -- CALL : R[A],...,R[A+C-2], R[A]( R[A+1], ..., R[A+B-1] )
       local fn = getreg(A)
       if type(fn)=="table" and fn.isfn==isfn then
          fn = fn.impl
          cipush(proto,pc,base,retpos,top,C-1,uvcache)
          proto = fn
          pc    = 1
          uvcache = {}
          local __top = B==0 and top or base+A+B
          base = base+A+1
          retpos = base - 1
          if proto.varflg>0 then
             local narg = top - base
             local nfix = proto.numpr
             for i=__top,base+nfix-1 do stk[i] = nil end
             -- fx1 fx2 .. fxk va1 .. van
             -- nil nil .. nil            fx1 fx2 .. fxk
             --                          Base           Top にする
             for i=0,nfix-1 do
               stk[__top+i] = stk[base+i]
               stk[base+i]  = nil
             end
             base  = __top
             __top = __top + nfix
          else
             if __top > base+proto.numpr then
               __top = base+proto.numpr
             end
          end
          top  = base + proto.maxstk
          for i=__top,top-1 do
            stk[i] = nil
          end
       else
          local function recv(...)
            if C==0 then
              local a = {...}
              setregs(A, #a, ...)
              top = base+A+#a
            else
              setregs(A, C-1, ...)
            end
          end
          recv(fn(getregs(A+1, B==0 and top-(base+A+1) or B-1)))
       end
    elseif OP == 29 then -- TAILCALL : return R[A]( R[A+1], ..., R[A+B-1] )
       closeupvals()
       local fn = getreg(A)
       if type(fn)=="table" and fn.isfn==isfn then
          fn = fn.impl
          proto = fn
          pc    = 1
          uvcache = {}
          local __top = B==0 and top or base+A+B
          for i=base+A,__top do
             stk[base-1 + i-(base+A)] = stk[i]
          end
          __top = base-1 + __top-(base+A)
          if proto.varflg>0 then
             local narg = top - base
             local nfix = proto.numpr
             for i=__top,base+nfix-1 do stk[i] = nil end
             -- fx1 fx2 .. fxk va1 .. van
             -- nil nil .. nil            fx1 fx2 .. fxk
             --                          Base           Top にする
             for i=0,nfix-1 do
               stk[__top+i] = stk[base+i]
               stk[base+i]  = nil
             end
             base  = __top
             __top = __top + nfix
          else
             if __top > base+proto.numpr then
               __top = base+proto.numpr
             end
          end
          top  = base + proto.maxstk
          for i=__top,top-1 do
            stk[i] = nil
          end
       else
          local function recv(...)
            local a = {...}
            local __base, __top
            proto,pc,__base,__retpos,__top,wanted,uvcache = cipop()
            if wanted == -1 then -- multival
              setregs(retpos-base, #a, ...)
              base,retpos,top = __base, __retpos, retpos+#a
            else
              setregs(retpos-base, wanted, ...)
              base,retpos,top = __base, __retpos, __top
            end
          end
          if citop == 1 then
            return fn(getregs(A+1, B==0 and top-(base+A+1) or B-1))
          else
            recv(fn(getregs(A+1, B==0 and top-(base+A+1) or B-1)))
          end
       end
    elseif OP == 30 then -- RETURN : R[A], ..., R[A+B-1]
       closeupvals()
       if citop == 1 then
          return
       else
         local __base, __top
         proto,pc,__base,__retpos,__top,wanted,uvcache = cipop()
         local numret = B==0 and top-(base+A) or B-1
         if wanted == -1 then -- multival
           setregs(retpos-base, numret, getregs(A,numret))
           base,retpos,top = __base, __retpos, retpos+numret
         else
           setregs(retpos-base, wanted, getregs(A,numret))
           base,retpos,top = __base, __retpos, __top
         end
       end
    elseif OP == 31 then -- FORLOOP for i=R[A],R[A+1],R[A+2] do end のループ継続条件
       local idx = getreg(A)+getreg(A+2)
       local continue
       if getreg(A+2)>0 then --uploop
          continue = (idx <= getreg(A+1))
       else
          continue = (idx >= getreg(A+1))
       end
       if continue then
         pc = pc + sBx
         setreg(A, idx)   -- ループ用の内部カウンタ
         setreg(A+3, idx) -- Luaのコードから見える変数
       end
    elseif OP == 32 then -- FORPREP for i=R[A],R[A+1],R[A+2] do end の開始。まず FORLOOP までジャンプ
       setreg(A, getreg(A)-getreg(A+2))
       pc = pc + sBx
-- 33 : TFORLOOP
-- vmLoopを再帰呼びしないといけないめんどいこまった
    elseif OP == 52 then -- TFORLOOPの直後のJMPを無理矢理実装 : pc += sBx
       A = bits(proto.code[pc-2], 6, 8) + 3
       if getreg(A) ~= nil then
         setreg(A-1, getreg(A))
         pc = pc + sBx
       end
    elseif OP == 33 then -- TFORLOOP for _ in R[A] do end のループ継続条件
                         -- R[A+1], R[A+2] に state と index
       setreg(A+5, getreg(A+2))
       setreg(A+4, getreg(A+1))
       setreg(A+3, getreg(A))
       A   = A + 3
       top = base+A+3
       B = 0
       C = C+1
       -- CALL HERE!!! --------------------------
       if bits(proto.code[pc],0,6) == 22 then
         proto.code[pc] = proto.code[pc]+30 -- 新ジャンル：動的命令書き換え JMP(22)->52
       end
       local fn = getreg(A)
       if type(fn)=="table" and fn.isfn==isfn then
          fn = fn.impl
          cipush(proto,pc,base,retpos,top,C-1,uvcache)
          proto = fn
          pc    = 1
          uvcache = {}
          local __top = B==0 and top or base+A+B
          base = base+A+1
          retpos = base - 1
          if proto.varflg>0 then
             local narg = top - base
             local nfix = proto.numpr
             for i=__top,base+nfix-1 do stk[i] = nil end
             -- fx1 fx2 .. fxk va1 .. van
             -- nil nil .. nil            fx1 fx2 .. fxk
             --                          Base           Top にする
             for i=0,nfix-1 do
               stk[__top+i] = stk[base+i]
               stk[base+i]  = nil
             end
             base  = __top
             __top = __top + nfix
          else
             if __top > base+proto.numpr then
               __top = base+proto.numpr
             end
          end
          top  = base + proto.maxstk
          for i=__top,top-1 do
            stk[i] = nil
          end
       else
          local function recv(...)
            if C==0 then
              local a = {...}
              setregs(A, #a, ...)
              top = base+A+#a
            else
              setregs(A, C-1, ...)
            end
          end
          recv(fn(getregs(A+1, B==0 and top-(base+A+1) or B-1)))
       end
       -- CALL HERE!!! --------------------------
       --if getreg(A) ~= nil then -- ループ継続
       --   pc = pc + xsBx() -- 次のジャンプ命令でジャンプバック
       --else
       --   pc = pc + 1
       --end
    elseif OP == 34 then -- SETLIST : RA[(C-1)*FPF+i] = R[A+i] for 1<=i<=B
       local FPF = 50 -- マクロ定義
       if B==0 then B = top-(base+A)-1 end
       if C==0 then C = proto.code[pc] pc=pc+1 end -- ひょええ
       local arr = getreg(A)
       for i=1,B do
          arr[(C-1)*FPF+i] = getreg(A+i)
       end
    elseif OP == 35 then -- CLOSE : R[A]以上の変数を
       closeupvals(A)
    elseif OP == 36 then -- CLOSURE : R[A] = Funcs[Bx]
       local function xOP() return bits(proto.code[pc], 0, 6) end
       local function xB()  return bits(proto.code[pc],23, 9) end

       local cl = toclosure(proto.kf[Bx+1])
       for i=1, cl.numup do
         if xOP() == 0 then -- MOVE : 現在のスタックの変数をupval化
            local b = xB()
            local uv = uvcache[b]
            if not uv then uv={base = stk, idx = base+b} uvcache[b]=uv end
            cl.upval[i] = uv
         else -- GETUPVAL : 親(祖先)のスタックのupvalを持ってくる
            cl.upval[i] = proto.upval[xB()+1]
         end
         pc = pc+1
       end
       setreg(A, {isfn=isfn, impl=cl})
    elseif OP == 37 then -- VARARG : R[A],...,R[A+B-2] = ...
       local numvar = base-retpos-1-proto.numpr
       if B == 0 then -- multivalue
          B = numvar
          top = base + A + numvar
       else
          B = B-1
       end
       for i=0,B-1 do
         if i < numvar then
           setreg(A+i, stk[base-numvar+i])
         else
           setreg(A+i, nil)
         end
       end
    end
  end
end

-------------------------------------------------------------------------
-- Main
-------------------------------------------------------------------------

arg[1] = arg[1] or "luac.out"
ld = loader(readFile(arg[1]))
main = ld:func()
vmLoop(main)
